<'
# Copyright (c) 2011-2015 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Feb 29 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]

lang c
'>
#include "autoconf/acheader.h"

#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>

#include "commonStructLib.h"
#include "h2evnLib.h"
#include "taskLib.h"

#include "<"$comp">_parameters.h"
#include "<"$comp">_msglib.h"
#include "<"$comp">_activity.h"
#include "<"$comp">_control_task.h"
#include "typecopy.h"


/* --- local data ---------------------------------------------------------- */

#define CNTRL_TASK_MIN_STACK_SIZE 4096
#define EXEC_TASK_MIN_STACK_SIZE  4096

/* cslib server with context pointer, for reentrant csLib callbacks */
struct CS_SERV_IDS {
  CS_SERV csserv;
  struct genom_component_data *self;
};

<'foreach s [$component services] {'>
static void	<"$comp">_<"[$s name]">_rqstcb(SERV_ID csserv, int sid);
<'}'>
static void *	<"$comp">_cntrl_task(void *);


/* <"[--- genom_${comp}_init --------------------------------------------]"> */

void *
genom_<"$comp">_init(void)
{
  struct genom_component_data *self;
  char tname[64];

  /* create internal data structure */
  self = malloc(sizeof(*self));
  if (!self) {
    genom_log_warn(1, "cannot create internal data structure");
    return NULL;
  }

<'if {![catch {$component ids} ids]} {'>
  genom_tinit_<"[$ids mangle]">(&self->ids);
<'}'>
<'foreach p [$component ports] {'>
  genom_tinit_<"$comp">_<"[$p name]">_port(&self->ports.<"[$p name]">);
<'}'>
<'foreach r [$component remotes] {'>
  genom_tinit_<"$comp">_<"[$r name]">_remote(&self->remotes.<"[$r name]">);
<'}'>
<'foreach t [$component tasks] { '>

  self->tasks.<"[$t name]">.context.raise = genom_pocolibs_raise;
  self->tasks.<"[$t name]">.context.raised = genom_pocolibs_raised;
  self->tasks.<"[$t name]">.context.data =
    &self->tasks.<"[$t name]">.context_data;

  self->tasks.<"[$t name]">.context_data.self = self;
  self->tasks.<"[$t name]">.context_data.ex = genom_ok;
  self->tasks.<"[$t name]">.context_data.exdetail = NULL;
  self->tasks.<"[$t name]">.context_data.exsize = 0;

  self->tasks.<"[$t name]">.taskid = ERROR;

  self->tasks.<"[$t name]">.rusage.cycles = 0;
  self->tasks.<"[$t name]">.rusage.timings.last = 0.;
  self->tasks.<"[$t name]">.rusage.timings.max = 0.;
  self->tasks.<"[$t name]">.rusage.timings.avg = 0.;
  self->tasks.<"[$t name]">.rusage.load.last = 0.;
  self->tasks.<"[$t name]">.rusage.load.max = 0.;
  self->tasks.<"[$t name]">.rusage.load.avg = 0.;
  pthread_spin_init(&self->tasks.<"[$t name]">.rlock, 0);

  self->tasks.<"[$t name]">.runnable = 1;
  self->tasks.<"[$t name]">.shutdown = 0;
  {
    int i;
    for(i = 0; i < genom_max_activities(); i++)
      self->tasks.<"[$t name]">.activities.a[i].any.status = ACT_VOID;
    i = genom_<"$comp">_activity_alloc(&self->tasks.<"[$t name]">.activities);
    assert(i == 0);
    self->tasks.<"[$t name]">.permanent =
      &self->tasks.<"[$t name]">.activities.a[i].any;
  }
  self->tasks.<"[$t name]">.permanent->sid = -1;
  self->tasks.<"[$t name]">.permanent->rid = -1;
  self->tasks.<"[$t name]">.permanent->status = ACT_INIT;
  self->tasks.<"[$t name]">.permanent->start = 1;

  pthread_mutex_init(&self->tasks.<"[$t name]">.lock, NULL);
  pthread_cond_init(&self->tasks.<"[$t name]">.sync, NULL);
<'}'>

  self->control.context.raise = genom_pocolibs_raise;
  self->control.context.raised = genom_pocolibs_raised;

  self->control.context.data = &self->control.context_data;
  self->control.context_data.self = self;
  self->control.context_data.ex = genom_ok;
  self->control.context_data.exdetail = NULL;
  self->control.context_data.exsize = 0;

  self->control.taskid = ERROR;
  snprintf(self->control.mbox_name, sizeof(self->control.mbox_name),
           "%s", genom_instance);
  self->control.csserv = NULL;
  self->control.shutdown = 0;
  self->control.activity.any.status = ACT_VOID;
  memset(self->control.run_map, 0, sizeof(self->control.run_map));
  pthread_spin_init(&self->control.rlock, 0);
  pthread_mutex_init(&self->control.lock, NULL);
  pthread_cond_init(&self->control.sync, NULL);

  self->resources.all = 0;
  self->resources.control = NULL;
<'foreach t [$component tasks] { '>
  self->resources.task_<"[$t name]"> = NULL;
<'}'>
  self->resources.q = 0;
  self->resources.qnext = 0;
  pthread_mutex_init(&self->resources.lock, NULL);
  pthread_cond_init(&self->resources.sync, NULL);

  /* create posters */
  if (genom_metadata_<"$comp">_init(self)) goto error;
  if (genom_state_<"$comp">_init(self)) goto error;
<'foreach p [$component ports simple] {'>
  if (genom_<"$comp">_<"[$p name]">_open(&self->control.context)) goto error;
<'}'>
  if (genom_<"$comp">_genom_metadata_write(&self->control.context)) goto error;
  if (genom_<"$comp">_genom_state_write(&self->control.context)) goto error;

<'foreach t [$component tasks] {'>
<'  if {[catch {[$t priority] value} priority]} { set priority 255 }'>
<'  if {[catch {[$t stack] value} stack]} { set stack [expr {1<<23}] }'>
  /* spawn <"[$t name]"> task */
  snprintf(tname, sizeof(tname), "%s/%s", genom_instance, "<"[$t name]">");
  pthread_mutex_lock(&self->tasks.<"[$t name]">.lock);
  self->tasks.<"[$t name]">.taskid =
    taskSpawn2(tname, <"$priority">,
               VX_FP_TASK, EXEC_TASK_MIN_STACK_SIZE + <"$stack">,
               genom_<"$comp">_<"[$t name]">_exec_task, self);
  while (self->tasks.<"[$t name]">.taskid != ERROR &&
         self->tasks.<"[$t name]">.permanent->start)
    pthread_cond_wait(&self->tasks.<"[$t name]">.sync,
                      &self->tasks.<"[$t name]">.lock);

  genom_state_<"$comp">_update(self, &self->tasks.<"[$t name]">.activities,
    <"${comp}_[$t name]_TASKID">);
  if (self->tasks.<"[$t name]">.taskid == ERROR) {
    genom_log_warn(1, "cannot spawn %s exec task", "<"[$t name]">");
    pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);
    goto error;
  }
  if (self->tasks.<"[$t name]">.permanent->status == ACT_ETHER &&
      self->tasks.<"[$t name]">.permanent->state != <"$comp">_ether) {
    genom_log_warn(0, "%s exec task failed", "<"[$t name]">");
    pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);
    goto error;
  }
  pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);
  genom_state_<"$comp">_refresh(self);

<'}'>

  /* spawn control task */
  snprintf(tname, sizeof(tname), "%s", genom_instance);
  pthread_mutex_lock(&self->control.lock);
  if (taskSpawn2(tname, 10 /* priority */,
                 VX_FP_TASK, <"$COMP">_MAX_RQST_SIZE + <"$COMP">_MAX_REPLY_SIZE
                 + CNTRL_TASK_MIN_STACK_SIZE /*size*/,
                 <"$comp">_cntrl_task, self) == ERROR) {
    genom_log_warn(1, "cannot spawn control task");
    pthread_mutex_unlock(&self->control.lock);
    goto error;
  }
  while (self->control.taskid == ERROR)
    pthread_cond_wait(&self->control.sync, &self->control.lock);
  if (self->control.taskid == 0) {
    self->control.taskid = ERROR;
    pthread_mutex_unlock(&self->control.lock);
    goto error;
  }
  pthread_mutex_unlock(&self->control.lock);

  genom_log_info("setup and running");
  return self;

error:
  genom_<"$comp">_fini(self);
  return NULL;
}


/* <"[--- genom_${comp}_fini --------------------------------------------]"> */

void
genom_<"$comp">_fini(void *data)
{
  struct genom_component_data *self = data;

  /* interrupt & wait for all regular activities */
<'foreach t [$component tasks] {'>
  if (self->tasks.<"[$t name]">.taskid != ERROR) {
    struct genom_activity *a;
    int delay;
    int id;

    genom_log_debug("interrupting task <"[$t name]"> activities");
    pthread_mutex_lock(&self->tasks.<"[$t name]">.lock);
    do {
      delay = 0;
      for(id = 0; id < genom_max_activities(); id++) {
        a = &self->tasks.<"[$t name]">.activities.a[id].any;
        if (a == self->tasks.<"[$t name]">.permanent) continue;

        if (a->status == ACT_INIT || a->status == ACT_RUN) {
          if (!a->stop) {
            a->stop = 1;
            a->interruptedby = "kill";
            if (!self->tasks.<"[$t name]">.runnable) {
              self->tasks.<"[$t name]">.runnable = 1;
              pthread_cond_broadcast(&self->tasks.<"[$t name]">.sync);
            }
          }
          delay = 1;
        } else if (a->status == ACT_STOP)
          delay = 1;
      }
      if (delay)
        pthread_cond_wait(&self->tasks.<"[$t name]">.sync,
                          &self->tasks.<"[$t name]">.lock);
    } while(delay);
    pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);
  }

<'}'>
  /* interrupt permanent activities */
<'foreach t [lreverse [$component tasks]] {'>
  if (self->tasks.<"[$t name]">.taskid != ERROR) {
    struct genom_activity *a;

    genom_log_debug("interrupting task <"[$t name]">");
    pthread_mutex_lock(&self->tasks.<"[$t name]">.lock);

    a = self->tasks.<"[$t name]">.permanent;
    if (!a) {
      /* activity was ether, realloc one */
      int i = genom_<"$comp">_activity_alloc(
        &self->tasks.<"[$t name]">.activities);
      assert(i>=0);
      a = &self->tasks.<"[$t name]">.activities.a[i].any;
      a->sid = a->rid = -1;
    }
    if (a->status != ACT_STOP) {
      if (a->status == ACT_ETHER) a->status = ACT_INIT;
      a->stop = 1;
      a->interruptedby = "kill";
      self->tasks.<"[$t name]">.permanent = a;
      self->tasks.<"[$t name]">.runnable = 1;
      pthread_cond_broadcast(&self->tasks.<"[$t name]">.sync);
    }
    while(self->tasks.<"[$t name]">.permanent &&
          self->tasks.<"[$t name]">.permanent->status != ACT_ETHER)
      pthread_cond_wait(&self->tasks.<"[$t name]">.sync,
                        &self->tasks.<"[$t name]">.lock);

    pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);
  }

<'}'>
  /* wait for exec tasks */
<'foreach t [lreverse [$component tasks]] {'>
  genom_log_debug("waiting task <"[$t name]">");
  pthread_mutex_lock(&self->tasks.<"[$t name]">.lock);
  self->tasks.<"[$t name]">.shutdown = 1;
  self->tasks.<"[$t name]">.runnable = 1;
  pthread_cond_broadcast(&self->tasks.<"[$t name]">.sync);
  while (self->tasks.<"[$t name]">.taskid != ERROR) {
    pthread_cond_wait(&self->tasks.<"[$t name]">.sync,
                      &self->tasks.<"[$t name]">.lock);
  }
  pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);

<'}'>

  /* wait for control task */
  if (self->control.taskid != ERROR) {
    genom_log_debug("waiting control task");
    pthread_mutex_lock(&self->control.lock);
    self->control.shutdown = 1;
    while (self->control.taskid != 0) {
      genom_<"$comp">_schedule_cntrl(self);
      pthread_cond_wait(&self->control.sync, &self->control.lock);
    }
    pthread_mutex_unlock(&self->control.lock);
  }

  /* clean up */
<'foreach p [$component ports] {'>
  genom_<"$comp">_<"[$p name]">_delete(&self->control.context);
<'}'>

<'foreach r [$component remotes] {'>
  genom_tfini_<"$comp">_<"[$r name]">_remote(&self->remotes.<"[$r name]">);
<'}'>
<'foreach p [$component ports] {'>
  genom_tfini_<"$comp">_<"[$p name]">_port(&self->ports.<"[$p name]">);
<'}'>
<'if {![catch {$component ids} ids]} {'>
  genom_tfini_<"[$ids mangle]">(&self->ids);
<'}'>
  free(self);
  genom_log_info("shutdown complete");
}


/* <"[--- genom_${comp}_schedule_cntrl ----------------------------------]"> */

void
genom_<"$comp">_schedule_cntrl(struct genom_component_data *self)
{
  if (self->control.taskid == ERROR) return;
  h2evnSignal(self->control.taskid);
}


/* <"[--- ${comp}_cntrl_task(void) --------------------------------------]"> */

static void *
<"$comp">_cntrl_task(void *data)
{
  struct genom_component_data *self = data;
  int e;

  /* create requests reception mailbox */
  if (csMboxInit(self->control.mbox_name, <"$COMP">_MBOX_RQST_SIZE, 0) != OK) {
    genom_log_warn(1, "cannot create server mailbox");
    pthread_mutex_lock(&self->control.lock);
    self->control.taskid = 0;
    pthread_cond_broadcast(&self->control.sync);
    pthread_mutex_unlock(&self->control.lock);
    return NULL;
  }

  /* configure as a server */
  if (csServInitN(
	<"$COMP">_MAX_RQST_SIZE,
	<"$COMP">_MAX_REPLY_SIZE >= <"$COMP">_MAX_INTERMED_REPLY_SIZE ?
	<"$COMP">_MAX_REPLY_SIZE : <"$COMP">_MAX_INTERMED_REPLY_SIZE,
	<"$COMP">_NRQSTID, &self->control.csserv) != OK) {
    genom_log_warn(1, "cannot initialize server mailbox");
    goto error;
  }
  /* fiddle with returned CS_SERV structure, to store ids for callbacks */
  self->control.csserv =
    realloc(self->control.csserv, sizeof(struct CS_SERV_IDS));
  ((struct CS_SERV_IDS *)self->control.csserv)->self = self;

  /* install services callbacks */
<'foreach s [$component services] {'>
  if (csServFuncInstall(self->control.csserv, <"$COMP">_<"[$s name]">_RQSTID,
                        (FUNCPTR)<"$comp">_<"[$s name]">_rqstcb) != OK) {
    genom_log_warn(1, "cannot serve service <"[$s name]">");
    goto error;
  }
<'}'>

  /* signal parent */
  genom_log_info("spawned control task");
  pthread_mutex_lock(&self->control.lock);
  self->control.taskid = taskIdSelf();
  pthread_cond_broadcast(&self->control.sync);
  pthread_mutex_unlock(&self->control.lock);

  do {
    if (self->control.shutdown) {
      /* the task must sleep with h2evnSusp(), so there is no way to atomically
       * check for the shutdown flag. So do not bother taking the mutex.
       * The genom_<"$comp">_fini() function will signal an h2evn, after
       * setting the shutdown flag, so it shouldn't exist any race. */
      break;
    }

    if (h2evnSusp(0) != TRUE) {
      genom_log_warn(1, "someone did something nasty");
      genom_log_warn(0, "aborting");
      abort();
    }
    genom_log_debug("control task wake up");

    /* read reception mailbox status */
    e = csMboxStatus(RCV_MBOX);
    if (e & RCV_MBOX)  {
      if (csServRqstExec(self->control.csserv) != OK) {
        genom_log_warn(1, "someone did something nasty");
        genom_log_warn(0, "aborting");
        abort();
      }
    }


    /* process pending activities */
<'foreach t [$component tasks] {'>
    pthread_mutex_lock(&self->tasks.<"[$t name]">.lock);
    {
      struct genom_activity *a;
      int id;

      for(id = 0; id < genom_max_activities(); id++) {
        a = &self->tasks.<"[$t name]">.activities.a[id].any;
        switch(a->status) {
          case ACT_VOID: case ACT_RUN: case ACT_STOP: continue;

          case ACT_INIT: {
            int delay;
            if (!a->start) {
              pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);
              delay = genom_<"$comp">_interrupt_reqd(self, a);
              pthread_mutex_lock(&self->tasks.<"[$t name]">.lock);
              if (!delay) {
                a->start = 1;
                if (!self->tasks.<"[$t name]">.runnable) {
                  self->tasks.<"[$t name]">.runnable = 1;
                  pthread_cond_broadcast(&self->tasks.<"[$t name]">.sync);
                }
              }
            }
            break;
          }

          case ACT_ETHER:
            if (a == self->tasks.<"[$t name]">.permanent)
              self->tasks.<"[$t name]">.permanent = NULL;
            else {
              pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);
              genom_<"$comp">_activity_report(self, a);
              pthread_mutex_lock(&self->tasks.<"[$t name]">.lock);
            }
            a->status = ACT_VOID;
            genom_state_<"$comp">_update(
              self, &self->tasks.<"[$t name]">.activities,
              <"${comp}_[$t name]_TASKID">);
            break;
        }
      }
    }
    pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);
<'}'>

    /* update state port */
    genom_state_<"$comp">_refresh(self);
  } while(1);

  genom_log_info("shutting down control task");

error:
  if (self->control.csserv) csServEnd(self->control.csserv);
  csMboxEnd();

  pthread_mutex_lock(&self->control.lock);
  self->control.taskid = 0;
  pthread_cond_broadcast(&self->control.sync);
  pthread_mutex_unlock(&self->control.lock);
  return NULL;
}


/* --- Service callbacks --------------------------------------------------- */

<'foreach s [$component services] {'>

static void
<"$comp">_<"[$s name]">_rqstcb(SERV_ID csserv, int rid)
{
  struct genom_component_data *self = ((struct CS_SERV_IDS *)csserv)->self;
  struct genom_<"$comp">_<"[$s name]">_activity *a;
  STATUS s;

  genom_log_debug("handling request for <"[$s name]">");

  /* get an activity slot */
<'  if {[catch {$s task} t]} {'>
  a = &self->control.activity.s_<"[$s name]">;
  assert(a->h.status == ACT_VOID);
  a->h.status = ACT_INIT;
<'  } else {'>
  pthread_mutex_lock(&self->tasks.<"[$t name]">.lock);
  int id = genom_<"$comp">_activity_alloc(&self->tasks.<"[$t name]">.activities);
  if (id < 0) {
    pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);
    s = csServReplySend(
      csserv, rid, FINAL_REPLY, ERROR, NULL, 0,
      genom_<"$comp">_genom_too_many_activities_encodex);
    if (s == ERROR) assert(!"unexpected failure");
    return;
  }
  a = &self->tasks.<"[$t name]">.activities.a[id].s_<"[$s name]">;
  pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);
<'  }'>
  a->h.sid = <"$COMP">_<"[$s name]">_RQSTID;
  a->h.rid = rid;
  genom_tinit_<"$comp">_<"[$s name]">_activity(a);

  /* decode input */
  s = csServRqstParamsGet(csserv, rid, (char *)&a->in, 0,
                          genom_<"$comp">_<"[$s name]">_decode);
  if (s == ERROR) {
    genom_log_warn(0, "invalid input for service %s", "<"[$s name]">");
    a->h.state = genom_serialization_id;
    a->h.exdetail = NULL;
    genom_<"$comp">_activity_report(self, &a->h);
    goto clean;
  }

  /* run control, validate codel and simple codels */
  a->h.state = genom_<"$comp">_<"[$s name]">_controlcb(self, a);

  /* wake up sleeping activities */
<'    foreach t [$component tasks] {'>
  pthread_mutex_lock(&self->tasks.<"[$t name]">.lock);
  if (!self->tasks.<"[$t name]">.runnable) {
    self->tasks.<"[$t name]">.runnable = 1;
    pthread_cond_broadcast(&self->tasks.<"[$t name]">.sync);
  }
  pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);
<'    }'>

  /* check control codels */
  if (a->h.state) {
    a->h.exdetail = (void *)self->control.context.raised(
      NULL, &self->control.context);
    genom_<"$comp">_activity_report(self, &a->h);
    goto clean;
  }

<'  if {[catch {$s task} t]} {'>
  /* interrupt incompatible activities */
  genom_<"$comp">_interrupt_reqd(self, &a->h);

  /* update after/before array */
  self->control.run_map[<"$COMP">_<"[$s name]">_RQSTID] = 1;

  /* send final reply */
  a->h.state = <"$comp">_ether;
  genom_<"$comp">_activity_report(self, &a->h);
  goto clean;

<'  } else {'>
  /* update state port */
  pthread_mutex_lock(&self->tasks.<"[$t name]">.lock);
  genom_state_<"$comp">_update(
              self, &self->tasks.<"[$t name]">.activities,
              <"${comp}_[$t name]_TASKID">);
  pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);

  /* send intermediate reply */
  s = csServReplySend(csserv, rid, INTERMED_REPLY, OK,
                      (char *)&a->h.aid, sizeof(int), NULL);
  if (s == ERROR) {
    genom_log_warn(1, "cannot acknowledge service %s", "<"[$s name]">");
  }

  return;
<'  }'>

clean:
  assert(a->h.status == ACT_INIT);
<'  if {![catch {$s task} t]} {'>
  pthread_mutex_lock(&self->tasks.<"[$t name]">.lock);
<'  }'>
  a->h.status = ACT_VOID;
<'  if {![catch {$s task} t]} {'>
  pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);
<'  }'>
}

<'}'>

/* eof */
